function Wout = computeWout(uData, dData, noiselevel, a, rho, s_in, s_fb, ...
    W, Win, Wfb, initialRunlength, sampleRunlength)

% An elementary implementation of the basic batch ESN learning algorithm.
% Computes ESN output weights from training data. For use within the
% demo package for stochastic gradient descent optimization of the global
% ESN parameters, as documented in the paper by Jaeger et al, Optimization and 
% Applications of Echo State Networks with Leaky Integrator Neurons. 
% Neural Networks, April 2007.

% Created by H. Jaeger February 21, 2007. For conditions of use, see main
% routine "stochOpt" in this package.


% INPUTS
% 
% uData: training input (time index = row index)
% dData: training output 
% noiselevel: amplitude of uniform state noise inserted into reservoir 
% a: leaking rate
% rho: spectrac radius of reservoir weight matrix
% s_in: scaling of input weights
% s_fb: scaling of output feedback weights
% W: reservoir weight matrix
% Win: input weight matrix
% Wfb: output feedback weight matrix
% initialRunlength: initial washout period
% sampleRunlength: number of data time points used for training
%     
% OUTPUT
% 
% Wout: output weight matrix


N = size(W,1);
uDim = size(uData,2);
dDim = size(dData,2);

linearOutputUnits = 1;

totalstate =  zeros(N+uDim+dDim,1);    
internalState = totalstate(1:N,1);
inWM = s_in * Win;
ofbWM = s_fb * Wfb;
intWM = rho * W;
retWM = (1-a) * eye(N);


stateCollectMat = zeros(sampleRunlength, N + uDim);
teachCollectMat = zeros(sampleRunlength, dDim);


for t = 1:initialRunlength + sampleRunlength 
    
    if uDim > 0
        in = uData(t,:)';  
    else in = [];
    end
    teach = dData(t,:)';   
    
    %write input into totalstate
    if uDim > 0
        totalstate(N+1:N+uDim) = in; 
    end
    %update totalstate except at input positions 

    if noiselevel == 0 | noiselevel == 0.0 
        internalState = retWM * internalState + f([intWM, inWM, ofbWM]*totalstate);  
    else
        internalState = retWM * internalState + f([intWM, inWM, ofbWM]*totalstate + ...
            noiselevel *  (rand(N,1)-0.5));
    end
    
    totalstate = [internalState; in; teach];    
    
    %collect states and results for later use in learning procedure
    if (t > initialRunlength)  
        collectIndex = t - initialRunlength;
        stateCollectMat(collectIndex,:) = [internalState' in']; %fill a row
        if linearOutputUnits
            teachCollectMat(collectIndex,:) = teach';
        else
            teachCollectMat(collectIndex,:) = (fInverse(teach))'; %fill a row
        end
    end
    
    
end


Wout = (pinv(stateCollectMat) * teachCollectMat)'; 


