function [expResultsTrain, expResultsTest, expResultsOutW] = ...
    plotarho(gridN, FigNr, W, Win, Wfb, plotNL, linearOutputUnits, uData, dData, ...
    a_PL_total, rho_PL_total, s_in_PL_total, s_fb_PL_total, ...
    initialRunlength, sampleRunlength, freeRunlength, plotRunlength, verbose)

% This function computes the train and test NRMSE and the mean absolute
% output weights of an ESN for values of the leaking rate a and spectral
% radius rho ranging on a grid within the unit square. As a side effect,
% two landscape plots of the training NRMSE (in a version with natural and 
% in another version with  logarithmical scaling) is produced, as well as a
% landscape plot of the mean absolute output weights. In each of these
% plots, the traces of the online adaptation of a and rho are inserted.
% When running the ESN for the various a / rho values, the values for s_in 
% and s_fb are fixed at the last value computed in the online adaptation.
% For further explanation see the paper by Jaeger et al, Optimization and 
% Applications of Echo State Networks with Leaky Integrator Neurons. 
% Neural Networks, April 2007 
%
% Created by H. Jaeger February 21, 2007. For conditions of use, see main
% routine "stochOpt" in this package.
%
% INPUTS
%
% gridN: an integer larger than 1, gives number of grid cells for the axes of 
%        the a-rho-landscape plot.
% FigNr: Matlab figure number in which the three plots appear
% W: ESN reservoir weight matrix (normalized to unit spectral radius)
% Win: input weight matrix
% Wfb: output feedback weight matrix
% plotNL: amplitude of uniform state noise inserted into ESN as a regularizer 
% linearOutputUnits: flag to indicate whether the ESN has linear (value 1) or
%                    sigmoid (value 0) output units
% uData: input Data; time index = row index. Column length must be at least   
%        initialRunlength + sampleRunlength + freeRunlength + plotRunlength
% dData: output Data (teacher)
% a_PL_total: a matrix of size N x T, where N is the number of values recorded for 
%             leaking rate a during the stoch. optimization and T is the number of 
%             tries (optimizations run simultaneously from different starting values)
% rho_PL_total: dito, for spectral radius rho
% s_in_PL_total: dito, for input weight scaling s_in
% s_fb_PL_total: dito, for output feedback weight scaling s_fb
% initialRunlength: initial washout lenght for the ESN run per each grid cell
% sampleRunlength: data "harvesting" length used for each cell
% freeRunlength: length of ESN simulation time after end of learning and 
%                before testing
% plotRunlength: number of ESN updates used for computing the test NRMSE
% verbose: binary flag set to 1 if textual output for monitoring progress is 
%          desired
%                    
% 
% OUTPUTS
% 
% expResultsTrain: matrix of size gridN x gridN containing the training NRMSEs for the 
%                  ESN learn/test runs for the various a and rho values. Left upper matrix 
%                  entry corresponds to a = 1/gridN, rho = 1/gridN. If the cell corresponds to 
%                  a setting of a and rho which yields an effective spectral radius exceeding 1, 
%                  a value of -0.1 appears in this matrix (and no learn/test run is executed).                 
% expResultsTest: dito, for the test NRMSEs
% expResultsOutW: dito, for the mean absolute output weights. 



expResultsTrain = -0.1*ones(gridN+1,gridN+1);
expResultsTest = expResultsTrain;
expResultsOutW = expResultsTrain;
expResultsSpecRad = expResultsTrain;

N = size(W,1);
s_in = mean(s_in_PL_total(end,:)); s_fb = mean(s_fb_PL_total(end,:));

for ii=0:gridN
    rho = (1/gridN) * ii;
    if verbose
        disp(sprintf('SR %g ', rho));
    end
    for jj=0:gridN
        a = (1/gridN) * jj;
        
        retainment = 1 - a ;
        retWM = diag(retainment * ones(1,N));
        intWM = rho * W;
        
        maxval = max(abs(myeigs(intWM + retWM)));
        
        
        retWM = sparse(retWM);        
        expResultsSpecRad(ii+1,jj+1) = maxval ;  
        
        if maxval > 1.0 
            expResultsTrain(ii+1,jj+1) = -0.01;
            expResultsTest(ii+1,jj+1) = -0.01;
            expResultsOutW(ii+1,jj+1) = -0.01;
            
        else      
            [NRMSEtrain, NRMSEtest, NRMSEtrainVec, NRMSEtestVec, WoutLearnt] = ...
                testESN(W, Win, Wfb,  plotNL, ...
                linearOutputUnits, a, rho, s_in, s_fb, uData, dData,...
                initialRunlength, sampleRunlength, freeRunlength, plotRunlength);            
            expResultsTrain(ii+1,jj+1) = NRMSEtrainVec;
            expResultsTest(ii+1,jj+1) = NRMSEtestVec;
            expResultsOutW(ii+1,jj+1)= mean(mean(abs(WoutLearnt)));
        end        
    end    
end

figure(FigNr); clf; set(FigNr, 'Position', [86   628   932   227]);

subplot(1,3,1);
plotTraces(expResultsTrain, min(min(expResultsTrain(expResultsTrain > 0))), ...
    max(max(expResultsTrain)), 1, a_PL_total, rho_PL_total, 0, 1:size(a_PL_total,2) );
title(sprintf('log train NRMSE'), 'FontSize', 7);

subplot(1,3,2);
plotTraces(expResultsTrain, min(min(expResultsTrain(expResultsTrain > 0))), ...
    max(max(expResultsTrain)), 0, a_PL_total, rho_PL_total, 0, 1:size(a_PL_total,2) );
title(sprintf('train NRMSE'), 'FontSize', 7);

subplot(1,3,3);
plotTraces(expResultsOutW, min(min(expResultsOutW(expResultsOutW > 0))), ...
    max(max(expResultsOutW)), 1, a_PL_total, rho_PL_total, 0, [] );
title(sprintf('log mean abs Wout'), 'FontSize', 7);
