%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Main Script for ESN parameter optimization by %%%%%%%%%%
%%%%%%%%%% stochastic gradient descent %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This script implements the stochastic gradient descent method for ESN
% paramater optimization documented in the paper: Herbert Jaeger, Mantas
% Lukosevicius, Dan Popovici, Optimization and Applications of Echo State
% Networks with Leaky Integrator Neurons. Neural Networks, April 2007 
% (preprint  available from first author's publications page)

% created and copyright: Herbert Jaeger, Feb 21, 2007

% This software is free for non-commercial use. If commercial use is
% intended, contact Fraunhofer IAIS (www.iais.fraunhofer.de) who have
% claimed international patents on ESN algorithms (pending).

% This software is intended for research use by experienced Matlab users 
% and includes no warranties or services. 
% Bug reports are gratefully received by h.jaeger@iu-bremen.de

% Basic working principle of this script: the online adaptation of the
% global parameters is run for a certain time (user-input-specified), after
% which the diagnostics are plotted and the user is asked for a
% continuation of the run. In each newly triggered continuation, the
% learning rates can be newly entered (let's call these intervals
% "user-query intervals")



%%%%%%%%%% setting control parameters for this script %%%%%%%%%%%%
newData = 1; % set to 1 if new training/testing data are to be generated, else 0 
dataType = 1; % indicator for which kind of data are to be generated. Set as follows: 
              % 1 superimposed Sines delay 5 (as in the Neural Networks
              % paper)
              % 2 whiteNoise delay 5 (delay 5 recall of white noise input)
              % 3 iterated tent map 
totalDataLength = 20000;    % length of training sequence. Is cyclically re-used 
                            % if online optimization is run longer 
blocklength = 500; % length of time segments after which a new point is plotted in  
                   % the graphical diagnostic outputs. Also the batch length of
                   % data which are "harvested" if the output weights are
                   % calculated not online by the LMS rule but by batch
                   % update
newReservoir = 1;  % set to 1 if a new reservoir is to be created, else 0
N = 20; % reservoir size
LMSforWout = 0; % set to 1 if the output weights are computed online by LMS
                % (stochastic gradient descent); set to 1 if output weights
                % are computed by batch update in intervals of
                % "blocklength"
freeze_s_in = 1; % set to 1 if the parameter s_in is not to be adapted. Is motivated by 
                 % making the graphical output (performance level plots as
                 % in the Neural Networks paper) more intelligible. If you
                 % want to get an optimized ESN, set to 0; then s_in will
                 % be optimized, too.  
newInitWout = 1; % Set to 1 if a new set of initial output weights is to be computed before 
                 % the adaptation of the global control parameters starts.
                 % Set to 0, a previously computed set of initial output
                 % weights is re-used (may be interesting with large ESNs
                 % to save CPU time)
initW0Noise = 0.001;  % The state noise level (range of uniform noise per reservoir
                       % neuron) used in computing the initial output
                       % weights
startValues = [0.33 0.2 0.3 0.0; 0.8 0.2 0.3 0.0;  0.8 0.66 0.3 0.0];
% startValues contains in each row the starting values 
% for a, rho, s_in, s_fb (in this order).
% If no entry for s_fb is given, it is assumed that there is no output
% feedback. 
% Notice: a is the leaking rate, rho the reservoir weight matrix spectral
% radius, s_in the scaling of the input weights, s_fb the scaling of the
% output feedback weights. 
% Example 1: startValues = [0.33 0.2 0.3 0.5; 0.8 0.2 0.3 0.5] means that the
% online adaptation of a, rho, s_in, s_fb is run simultaneously from two
% starting values, the first of which sets a = 0.33, rho = 0.2, s_in = 0.3,
% s_fb = 0.5. 
% Example 2: startValues = [0.33 0.2 0.3] means that the online adaptation
% is run for only a single starting setting with a = 0.33, rho = 0.2, s_in
% = 0.3. Because no entry for s_fb is given, the procedure will assume that
% no feedback weights are part of the model and keep s_fb = 0 throughout.
initialRunlength = 100; % the washout used for offline weight computation 
                        % (for the initial output weights, and for the
                        % greyshade computation in the performance level
                        % plots)
sampleRunlength = 500;  % the sampling length for offline weight computations
freeRunlength = 0;      % the timespan between ending the sampling/training and 
                        % beginning the testing. Not relevant in this
                        % script, set to 0.
plotRunlength = 100;    % the test data length used to compute testing error for 
                        % a test performance landscape similar to the
                        % training error landscapes from the Neural
                        % Networks paper. Not exploited in this script.
verbose = 1; % set to 1 if you want to have monitoring text output while the (time-intensive) 
             % computations of the performance landscape plots run.
             
%%%%%%%% initializing %%%%%%%%%%%%%%%%%%      

NrTries = size(startValues, 1); 
if size(startValues,2) == 3
    a = startValues(:,1)'; rho = startValues(:,2)'; s_in = startValues(:,3)'; 
    s_fb = 0; withFB = 0; NrParams = 3;
else 
    a = startValues(:,1)'; rho = startValues(:,2)'; s_in = startValues(:,3)'; 
    s_fb = startValues(:,4)'; withFB = 1; NrParams = 4;
end


%%%%%%%% create training data. 

% uData: input data, time index = row index. dData: output data.
if newData 
    if dataType == 1 
        T = 1:totalDataLength + 100;
        T = T';
        uData = sin(T) + sin(0.51 * T)  + sin(0.22 * T) + sin(0.1002 * T) + sin(0.05343 * T);
        uData = normalizeData01(uData);
        dData = uData(15:totalDataLength+14, 1);
        uData = uData(20:totalDataLength+19, 1);         
    elseif dataType == 2 
        uData = rand(totalDataLength + 100, 1);
        uData = normalizeData01(uData);
        dData = uData(15:totalDataLength+14, 1);        
        uData = uData(20:totalDataLength+19, 1);      
    elseif dataType == 3
        [uData, dData] = generateDataTentMap(totalDataLength);    
    end
end

dVar = var(dData); uDim = size(uData,2); dDim = size(dData,2); uZeros = zeros(uDim, NrTries);

if dDim > 1
    error('Error: current implementation of stochOpt only covers 1-dim output');
end

%%%%%%%% create new reservoir
if newReservoir
    connectivity = min([10/N 1]);
    W =  generate_internal_weights(N, connectivity);        
    Win = 2.0 * rand(N, uDim)- 1.0;     
    Wfb = rand(N,dDim)-0.5;
    if withFB
        repWfb = repmat(Wfb,1,NrTries);
    end
end

%%%%%%%% initialize state and output 
x = zeros(N, NrTries); % internal state of R1
y = zeros(dDim, NrTries);


%%%%%%%% initialize output weights (in matrix WoutInit)

if newInitWout
    WoutInit = zeros(NrTries, N + uDim);
    for nt = 1: NrTries
        if withFB
            WoutInit(nt,:) = ...
                computeWout(uData, dData, initW0Noise, startValues(nt,1), startValues(nt,2), ...
                startValues(nt,3), startValues(nt,4), W, Win, Wfb, 200, 1000);
        else
            WoutInit(nt,:) = ...
                computeWout(uData, dData, initW0Noise, startValues(nt,1), startValues(nt,2), ...
                startValues(nt,3), 0, W, Win, Wfb, 200, 1000);
        end
    end
end

%%%%%% set output weights Wout to initial weights
Wout = WoutInit;

%%%%%% initialize state gradients w.r.t. global parameters 
grad_xa = zeros(N,NrTries); grad_xrho = zeros(N,NrTries); 
grad_xs_in = zeros(N,NrTries); grad_xs_fb = zeros(N,NrTries);
grad_xa_old = zeros(N,NrTries); grad_xrho_old = zeros(N,NrTries); 
grad_xs_in_old = zeros(N,NrTries); grad_xs_fb_old = zeros(N,NrTries);

%%%%%% initialize block-internal collector matrices
y_BlockCollect = zeros(blocklength, dDim);
d_BlockCollect = zeros(blocklength, dDim);
u_BlockCollect = zeros(blocklength, uDim);
x_BlockCollect = zeros(blocklength, N, NrTries);
err_BlockCollect = zeros(blocklength, NrTries);
grad_a_BlockCollect = zeros(blocklength, NrTries);
grad_rho_BlockCollect = zeros(blocklength, NrTries);
grad_s_in_BlockCollect = zeros(blocklength, NrTries);
grad_s_fb_BlockCollect = zeros(blocklength, NrTries);

%%%%%% initialize total diagnostics collectors. 
% Convention: "PL" stands for  "Plot List", a list in which info is 
% harvested for later plotting outputs. The #_total plotlists collect the
% diagnostics across subsequent user-query intervals, and are appended with
% new diagnostic data after each user-query interval.

y_PL_total = zeros(0, NrTries);
d_PL_total = zeros(0, 1);
x_PL_total = zeros(0, N);
a_PL_total = zeros(0, NrTries);
rho_PL_total = zeros(0, NrTries);
s_in_PL_total = zeros(0, NrTries);
s_fb_PL_total = zeros(0, NrTries);
absgrad_a_PL_total = zeros(0, NrTries);
absgrad_rho_PL_total = zeros(0, NrTries);
absgrad_s_in_PL_total = zeros(0, NrTries);
absgrad_s_fb_PL_total = zeros(0, NrTries);
grad_a_PL_total = zeros(0, NrTries);
grad_rho_PL_total = zeros(0, NrTries);
grad_s_in_PL_total = zeros(0, NrTries);
grad_s_fb_PL_total = zeros(0, NrTries);
NRMSE_PL_total = zeros(0, NrTries);
avabs_Wout_PL_total = zeros(0, NrTries);

totalTime = 0; % counter of total network state updates elapsed 
goOn = 1; % flag whether the current user query interval should be the terminating one
          % (then this flag is set to 0 by user input)

while goOn
    if LMSforWout % case we have LMS update of output weights
        % here the user is asked to provide the specification for
        % continuing the online adaptation. The specification has to be
        % entered in the format it is queried, that is, a list of numbers
        % in "[]". The following items are queried:
        % runlength: for how many steps the next continuation will run
        % [LRs_Wout]: a list of learnrates for the LMS update of the output
        %             weights, one learnrate for each of the NrTries
        %             simultaneous adaptations
        % [LRs_para]: similarly, the learn rates for the global parameter
        %             stochastic gradient descent
        % FigNr: set to 0 if at the end of this continuation no performance
        %        landscape plots should be computed (they are
        %        time-expensive). If you want that graphics, set to the
        %        Matlab figure Nr where you want the graphics to be
        %        displayed.
        % gridNr: if performance landscape plotting is done, this number 
        %         specifies the gridsize. I use 8 for quick plots and 40
        %         for the fine-resolving plots from the Neural Networks
        %         paper.
        % NL: If performance landscape plottin is done, this is the
        %     regularizing state noiselevel that is inserted into the reservoir
        %     when computing the landscape cells.
        % goOn: set to 0 if the next continuation should be the last one, 
        %       else to 1.
        userIn = input('[runlength [LRs_Wout] [LRs_para] FigNr gridN NL goOn]: \n');
        runlength = userIn(1); LR_Wout = userIn(2:2+NrTries-1); 
        LR_para = userIn(3+NrTries-1:3+2*NrTries-2);  FigNr = userIn(4+2*NrTries-2); 
        gridN = userIn(5+2*NrTries-2); 
        NL = userIn(6+2*NrTries-2); goOn = userIn(7+2*NrTries-2); 
    else
        userIn = input('[runlength [LRs_para] FigNr gridN NL goOn]: \n');
        runlength = userIn(1);  LR_para = userIn(2:2+NrTries-1);  
        FigNr = userIn(3+NrTries-1); 
        gridN = userIn(4+NrTries-1); NL = userIn(5+NrTries-1); goOn = userIn(6+NrTries-1); 
        LR_Wout = NaN;
    end
    
    
    %%%%%% initialize diagnostics collectors for this  run continuation
    y_PL = zeros(ceil(runlength / blocklength), NrTries);
    d_PL = zeros(ceil(runlength / blocklength), 1);    
    a_PL = zeros(ceil(runlength / blocklength), NrTries);
    grad_a_PL = zeros(ceil(runlength / blocklength), NrTries);
    grad_rho_PL = zeros(ceil(runlength / blocklength), NrTries);
    grad_s_in_PL = zeros(ceil(runlength / blocklength), NrTries);
    grad_s_fb_PL = zeros(ceil(runlength / blocklength), NrTries);
    absgrad_a_PL = zeros(ceil(runlength / blocklength), NrTries);
    absgrad_rho_PL = zeros(ceil(runlength / blocklength), NrTries);
    absgrad_s_in_PL = zeros(ceil(runlength / blocklength), NrTries);
    absgrad_s_fb_PL = zeros(ceil(runlength / blocklength), NrTries);
    rho_PL = zeros(ceil(runlength / blocklength), NrTries);
    s_in_PL = zeros(ceil(runlength / blocklength), NrTries);
    s_fb_PL = zeros(ceil(runlength / blocklength), NrTries);
    NRMSE_PL = zeros(ceil(runlength / blocklength), NrTries);
    avabs_Wout_PL = zeros(ceil(runlength / blocklength), NrTries);
    
    %%%%%% run this user query interval
    
    blockCounter = 0;
    
    for i = 1:runlength 
        %% update counters        
        totalTime = totalTime + 1;
        % blockCounter counts current block (used for plotting points only at times
        % n * blocklength) 
        if mod(i, blocklength) == 1 % at this time, the block counter is incremented
            blockIncFlag = 1; 
            blockCounter = blockCounter + 1;
            withinBlockCounter = 1;
        else
            blockIncFlag = 0;
            withinBlockCounter = withinBlockCounter + 1;
        end
        % cyclicTime cycles through external input sequence, which is
        % re-used when learning lasts longer than input was pre-generated
        cyclicTime = mod(totalTime, totalDataLength); 
        if cyclicTime == 0
            cyclicTime = totalDataLength;
        end        
        
        % read external input and teacher
        u_new = uData(cyclicTime,:)';         
        d_new = dData(cyclicTime,1) * ones(1,NrTries);  % NrTries copies of teacher d in a row
        if not(LMSforWout)
            u_BlockCollect(withinBlockCounter, :) = u_new';
            d_BlockCollect(withinBlockCounter, :) = d_new(1);
        end       
        
        % update reservoir state x  
        Wx = W * x;
        Winu_new = Win * u_new;
        % in order to catch all tries with single matrix updates (saving us
        % slow-running loop constructs) we here and later collect info
        % from/for all the simultaneous tries in one matrix. The matrix
        % constructions we have to do are much cheaper than loops which we
        % would have to do otherwise.
        Winu_new = repmat(Winu_new, 1, NrTries);
        if withFB
            excitation = Winu_new * diag(s_in) + Wx * diag(rho)  +   Wfb * y * diag(s_fb);
        else
            excitation = Winu_new * diag(s_in) + Wx * diag(rho);
        end
        x_new = x * diag(1-a) + tanh(excitation) ;
        
        if LMSforWout & i > runlength - blocklength
            % when Wout is computed with LMS, 
            % collect x only for plotting purposes and only for first try
            % and only for last visit of block
            x_BlockCollect(i - runlength + blocklength, :, 1) = x_new(:,1)';
        else
            % when Wout is computed with batch method, we need to collect
            % all reservoir states at any time
            x_BlockCollect(withinBlockCounter, :, :) = x_new ;
        end
        % update y and error
        xu_new = [x_new; repmat(u_new, 1, NrTries)]; 
        y_new = sum(Wout' .* xu_new); % the new y's for the tries in a row        
        err = d_new - y_new;
        err_BlockCollect(withinBlockCounter, :) = err;        
        
        %% update a, rho, s_in
        % compute state gradient w.r.t. parameters
        expPlus2x = exp(2*excitation); expMinus2x = exp(-2*excitation);
        tanhDerivative = 4 ./ (2 + expPlus2x + expMinus2x);
        if withFB             
            grad_xa_new = grad_xa * diag(1-a) - x + tanhDerivative .* ...
                (W * grad_xa * diag(rho) + Wfb * sum(Wout' .* [grad_xa_old; uZeros]) * diag(s_fb));
            grad_xrho_new = grad_xrho  * diag(1-a) + tanhDerivative .* ...
                (W * grad_xrho * diag(rho) + Wx + ...
                Wfb * sum(Wout' .* [grad_xrho_old; uZeros]) * diag(s_fb));
            grad_xs_in_new = grad_xs_in  * diag(1-a) + tanhDerivative .* ...
                (Winu_new + W * grad_xs_in * diag(rho) + ...
                Wfb * sum(Wout' .*  [grad_xs_in_old; uZeros]) * diag(s_fb));
            grad_xs_fb_new = grad_xs_fb  * diag(1-a) + tanhDerivative .* ...
                (W * grad_xs_fb * diag(rho) + ...
                Wfb * sum(Wout' .*  [grad_xs_fb_old; uZeros]) * diag(s_fb) + Wfb * y);
        else
            grad_xa_new = grad_xa * diag(1-a) - x + tanhDerivative .* (W * grad_xa * diag(rho));
            grad_xrho_new = grad_xrho * diag(1-a) + tanhDerivative .* (W * grad_xrho * diag(rho) + Wx);
            grad_xs_in_new = grad_xs_in * diag(1-a) + ...
                tanhDerivative .* (Winu_new + W * grad_xs_in * diag(rho));        
            grad_xs_fb_new = zeros(N, NrTries);
        end
        
        % compute error gradients 
        grad_a = - err .* sum(Wout' .* [grad_xa_new; uZeros]);
        grad_rho = - err .* sum(Wout' .* [grad_xrho_new; uZeros]);
        grad_s_in = - err .* sum(Wout' .* [grad_xs_in_new; uZeros]);
        grad_s_fb = - err .* sum(Wout' .* [grad_xs_fb_new; uZeros]);
        grad_a_BlockCollect(withinBlockCounter, :) = grad_a;
        grad_rho_BlockCollect(withinBlockCounter, :) = grad_rho;
        grad_s_in_BlockCollect(withinBlockCounter, :) = grad_s_in;
        grad_s_fb_BlockCollect(withinBlockCounter, :) = grad_s_fb;
        
        % update global parameters
        a = a - LR_para .* grad_a;
        rho = rho - LR_para .* grad_rho;
        if not(freeze_s_in)
            s_in = s_in - LR_para .* grad_s_in;   
        end
        s_fb = s_fb - LR_para .* grad_s_fb;
        % push parameters back to admissible range when necessary
        a(a > 1) = 1; 
        a(a < 0) = 0;   
        rho(rho < 0) = 0;
        s_in(s_in < 0) = 0.01;
        
        %% update output weights 
        if LMSforWout
            Wout = Wout + diag(LR_Wout .* err) * (xu_new' + NL * (rand(NrTries, N+uDim)));
        elseif blockIncFlag & i ~= 1
            for nt = 1:NrTries
                NoiseMat = NL * (rand(blocklength, N) - 0.5);
                Wout(nt,:) = ...
                    (pinv([x_BlockCollect(:,:,nt)+ NoiseMat, u_BlockCollect]) * d_BlockCollect)';
            end
        end        
        
        % at end of a block, compute diagnostics and insert them into the
        % plotlists for the this user query interval
        if blockIncFlag
            avabs_Wout_PL(blockCounter,:) = mean(abs(Wout'));
            d_PL(blockCounter,1) = d_new(1);
            y_PL(blockCounter,:) = y_new;
            NRMSE = sqrt(mean(err_BlockCollect .^ 2) / dVar);           
            NRMSE_PL(blockCounter,:) = NRMSE;
            grad_a_PL(blockCounter, :) = mean(grad_a_BlockCollect);
            grad_rho_PL(blockCounter, :) = mean(grad_rho_BlockCollect);
            grad_s_in_PL(blockCounter, :) = mean(grad_s_in_BlockCollect);
            grad_s_fb_PL(blockCounter, :) = mean(grad_s_fb_BlockCollect);
            absgrad_a_PL(blockCounter, :) = mean(abs(grad_a_BlockCollect));
            absgrad_rho_PL(blockCounter, :) = mean(abs(grad_rho_BlockCollect));
            absgrad_s_in_PL(blockCounter, :) = mean(abs(grad_s_in_BlockCollect));
            a_PL(blockCounter,:) = a;
            rho_PL(blockCounter,:) = rho;
            s_in_PL(blockCounter,:) = s_in;
            s_fb_PL(blockCounter,:) = s_fb;
        end        
        
        
        % assign new state values and gradients      
        x = x_new; y = y_new; 
        grad_xa_old = grad_xa; grad_xrho_old = grad_xrho; 
        grad_xs_in_old = grad_xs_in;   grad_xs_fb_old = grad_xs_fb;  
        grad_xa = grad_xa_new; grad_xrho = grad_xrho_new; 
        grad_xs_in = grad_xs_in_new; grad_xs_fb = grad_xs_fb_new;        
    end 
    
    %%%%%%% compute and display diagnostics
    
    % append new diagnostic plotlists to old
    y_PL_total = [y_PL_total; y_PL];
    d_PL_total = [d_PL_total; d_PL];    
    a_PL_total = [a_PL_total; a_PL];
    rho_PL_total = [rho_PL_total; rho_PL];
    s_in_PL_total = [s_in_PL_total; s_in_PL];
    s_fb_PL_total = [s_fb_PL_total; s_fb_PL];
    grad_a_PL_total = [grad_a_PL_total; grad_a_PL];
    grad_rho_PL_total = [grad_rho_PL_total; grad_rho_PL];
    grad_s_in_PL_total = [grad_s_in_PL_total; grad_s_in_PL];
    grad_s_fb_PL_total = [grad_s_fb_PL_total; grad_s_fb_PL];
    absgrad_a_PL_total = [absgrad_a_PL_total; absgrad_a_PL];
    absgrad_rho_PL_total = [absgrad_rho_PL_total; absgrad_rho_PL];
    absgrad_s_in_PL_total = [absgrad_s_in_PL_total; absgrad_s_in_PL];
    absgrad_s_fb_PL_total = [absgrad_s_fb_PL_total; absgrad_s_fb_PL];
    NRMSE_PL_total = [NRMSE_PL_total;  NRMSE_PL];
    avabs_Wout_PL_total = [avabs_Wout_PL_total; avabs_Wout_PL];
    
    plotTimes = 1:blocklength:blocklength*size(a_PL_total,1);
    
    % some reservoir state components for the first try and the last
    % recorded block
    figure(1); clf;
    set(1, 'Position', [76   698   384   249]); 
    for k = 1:4
        subplot(2,2,k);
        plot(x_BlockCollect(:,k)');
        if k == 1
            title('states from x, first try');
        end                       
    end
    
    % y vs. d
    figure(2); clf; 
    set(2, 'Position', [465   761   394   186]); 
    plot(plotTimes, d_PL_total, 'k', plotTimes, y_PL_total);            
    title('teacher(k) vs. predicted outputs(colors)');        
    
    % Normalized root mean square errors (NRMSEs) for the evolving models 
    figure(3); clf; set(3, 'Position', [868   748   408   201]); 
    plot(plotTimes, NRMSE_PL_total); 
    set(gca, 'YLim',  [0 1.5]);
    title('NRMSEs');
    
    % average absolute output weights
    figure(4); clf; set(4, 'Position', [465   519   398   162]); 
    plot(plotTimes, avabs_Wout_PL_total);
    title('av abs Wout');        
    
    % development of global parameters
    figure(5); clf; set(5, 'Position', [868    75   408   593]); 
    if withFB
        subplot(4,1,1);
        plot(plotTimes, a_PL_total);
        title('a'); 
        subplot(4,1,2);
        plot(plotTimes, rho_PL_total);
        title('rho'); 
        subplot(4,1,3);
        plot(plotTimes, s_in_PL_total);
        title('s\_in');
        subplot(4,1,4);
        plot(plotTimes, s_fb_PL_total);
        title('s\_fb');
    else
        subplot(3,1,1);
        plot(plotTimes, a_PL_total);
        title('a'); 
        subplot(3,1,2);
        plot(plotTimes, rho_PL_total);
        title('rho'); 
        subplot(3,1,3);
        plot(plotTimes, s_in_PL_total);
        title('s\_in'); 
    end
    
    % the absolute error gradients of the global parameters (blockwise
    % averages)
    figure(6); clf;  set(6, 'Position', [467    75   393   140]);
    if withFB
        subplot(4,1,1);
        plot(plotTimes, absgrad_a_PL_total);
        title('mean abs grads a');
        subplot(4,1,2);
        plot(plotTimes, absgrad_rho_PL_total);
        title('mean abs grads rho');
        subplot(4,1,3);
        plot(plotTimes, absgrad_s_in_PL_total);
        title('mean abs grads s\_in');
        subplot(4,1,4);
        plot(plotTimes, absgrad_s_fb_PL_total);
        title('mean abs grads s\_fb');
    else
        subplot(3,1,1);
        plot(plotTimes, absgrad_a_PL_total);
        title('mean abs grads a');
        subplot(3,1,2);
        plot(plotTimes, absgrad_rho_PL_total);
        title('mean abs grads rho');
        subplot(3,1,3);
        plot(plotTimes, absgrad_s_in_PL_total);
        title('mean abs grads s\_in');
    end
    
    % the error gradients of the global parameters (blockwise
    % averages)
    figure(7); clf;  set(7, 'Position', [459   295   402   148]);
    if withFB
        subplot(4,1,1);
        plot(plotTimes, grad_a_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads a');
        subplot(4,1,2);
        plot(plotTimes, grad_rho_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads rho');
        subplot(4,1,3);
        plot(plotTimes, grad_s_in_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads s\_in');
        subplot(4,1,4);
        plot(plotTimes, grad_s_fb_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads s\_fb');
    else
        subplot(3,1,1);
        plot(plotTimes, grad_a_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads a');
        subplot(3,1,2);
        plot(plotTimes, grad_rho_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads rho');
        subplot(3,1,3);
        plot(plotTimes, grad_s_in_PL_total, plotTimes, zeros(size(grad_a_PL_total,1),1), 'k');
        title('mean grads s\_in');
    end
    
    % this is the expensive optional computation of the performance
    % landscape plots. For your own diagnostic computations, the call to
    % "plotarho" returns the landscape values for training NRMSE, testing
    % NRMSE, and mean absolute output weight size in matrices resultsTrain,
    % resultsTest, resultsOutW. 
    if FigNr > 0
        linearOutputUnits = 1;        
        [resultsTrain, resultsTest, resultsOutW ] =... 
            plotarho(gridN, FigNr, W, Win, Wfb, NL, linearOutputUnits, uData, dData, ...
            a_PL_total, rho_PL_total, s_in_PL_total, s_fb_PL_total, ...
            initialRunlength, sampleRunlength, freeRunlength, plotRunlength, verbose);          
    end    
end
