function [NRMSEtrain, NRMSEtest, NRMSEtrainVec, NRMSEtestVec, WoutLearnt] = ...
    testESN(W, Win, Wfb,  noiselevel, linearOutputUnits, a, rho, s_in, s_fb, uData, dData, ...
    initialRunlength, sampleRunlength, freeRunlength, plotRunlength)

% This function trains an ESN on training data and computes various performance 
% indicators. For use within the
% demo package for stochastic gradient descent optimization of the global
% ESN parameters, as documented in the paper by Jaeger et al, Optimization and 
% Applications of Echo State Networks with Leaky Integrator Neurons. 
% Neural Networks, April 2007.

% Created by H. Jaeger February 21, 2007. For conditions of use, see main
% routine "stochOpt" in this package.


% INPUTS
% 
% W: reservoir weight matrix
% Win: input weight matrix
% Wfb: output feedback weight matrix
% noiselevel: amplitude of uniform state noise inserted into reservoir 
% linearOutputUnits: flag to indicate whether the ESN has linear (value 1) or
%                    sigmoid (value 0) output units
% a: leaking rate
% rho: spectrac radius of reservoir weight matrix
% s_in: scaling of input weights
% s_fb: scaling of output feedback weights
% uData: training input (time index = row index)
% dData: training output 
% initialRunlength: initial washout period
% sampleRunlength: number of data time points used for trainingTo be used 
% freeRunlength: length of ESN simulation time after end of learning and 
%                before testing
% plotRunlength: number of ESN updates used for computing the test NRMSE
% 
% OUTPUTS
% 
% NRMSEtrain: training NRMSE, a row vector with one NRMSE value per output dimension
% NRMSEtest: testing NRMSE, a row vector with one NRMSE value per output dimension
% NRMSEtrainVec: training NRMSE, a single scalar, giving the mean (across time) of 
%                squared vector norm of difference between correct and predicted 
%                output. For 1-dim output, same as NRMSEtrain
% NRMSEtestVec: dito, for test NRMSE
% WoutLearnt: learnt output weight matrix


N = size(W,1); uDim = size(uData,2); dDim = size(dData,2);
totalstate =  zeros(N+uDim+dDim,1); 
internalState = totalstate(1:N,1);
intWM = rho * W;
inWM = s_in * Win;
ofbWM = s_fb * Wfb;
WoutLearnt = zeros(dDim, N + uDim);
retWM = diag((1-a) * ones(1,N));
dVar = var(dData);
dDataNormalized = dData - repmat(mean(dData), size(dData,1), 1);
dVarVec = mean(sum((dDataNormalized.^2)'));

stateCollectMat = zeros(sampleRunlength, N + uDim);
teachCollectMat = zeros(sampleRunlength, dDim);
msetest = zeros(1,dDim); % the test mse's individually per output dimension
msetestVec = 0; % the test mse of vector-output difference norm
msetrain = zeros(1,dDim); % the train mse's individually per output dimension
msetrainVec = 0; % the train mse of vector-output difference norm


for t = 1:initialRunlength + sampleRunlength + freeRunlength + plotRunlength 
    
    if uDim > 0
        in = uData(t,:)';  
    else in = [];
    end
    teach = dData(t,:)';   
    
    %write input into totalstate
    if uDim > 0
    totalstate(N+1:N+uDim) = in; 
end
    %update totalstate except at input positions 
    if noiselevel == 0 | noiselevel == 0.0 | t > initialRunlength + sampleRunlength
        internalState = retWM * internalState + f([intWM, inWM, ofbWM]*totalstate);  
    else
        internalState = retWM * internalState + f([intWM, inWM, ofbWM]*totalstate + ...
            noiselevel * (rand(N,1)-0.5));
    end
    if linearOutputUnits
        netOut = WoutLearnt *[internalState;in];
    else
        netOut = timeConstant * f(WoutLearnt *[internalState;in]);
    end
    totalstate = [internalState;in;netOut];    
    
    %collect states and results for later use in learning procedure
    if (t > initialRunlength) & (t <= initialRunlength + sampleRunlength) 
        collectIndex = t - initialRunlength;
        stateCollectMat(collectIndex,:) = [internalState' in']; %fill a row
        if linearOutputUnits
            teachCollectMat(collectIndex,:) = teach';
        else
            teachCollectMat(collectIndex,:) = (fInverse(teach))'; %fill a row
        end
    end
    %force teacher output 
    if t <= initialRunlength + sampleRunlength
        totalstate(N+uDim+1:N+uDim+dDim) = teach' ; 
    end
    %update msetest
    if t > initialRunlength + sampleRunlength + freeRunlength
        for j = 1:dDim
            msetest(1,j) = msetest(1,j) + (teach(j,1)- netOut(j,1))^2;
       end
       msetestVec = msetestVec + (teach - netOut)'  * (teach - netOut);
    end
    %compute new model
    if t == initialRunlength + sampleRunlength
        WoutLearnt = (pinv(stateCollectMat) * teachCollectMat)'; 
        %compute mean square errors on the training data using the newly
        %computed weights
        for j = 1:dDim
            if linearOutputUnits
                msetrain(1,j) = sum((teachCollectMat(:,j) - ...
                    (stateCollectMat * WoutLearnt(j,:)')).^2);                
            else
                msetrain(1,j) = sum( (f(teachCollectMat(:,j)) - ...
                    f(stateCollectMat * WoutLearnt(j,:)')).^2);
            end
            msetrain(1,j) = msetrain(1,j) / sampleRunlength;
        end
        if linearOutputUnits
            errs = teachCollectMat - stateCollectMat * WoutLearnt';
            msetrainVec = mean(sum((errs.^2)'));            
        else
            errs = f(teachCollectMat) - f(stateCollectMat * WoutLearnt');
            msetrainVec = mean(sum((errs.^2)'));  
        end
    end    
    
end
%end of the great do-loop


 msetestresult = msetest / plotRunlength;
 msetestVecresult = msetestVec / plotRunlength;
 NRMSEtrain = sqrt(msetrain ./ dVar);
 NRMSEtest = sqrt(msetestresult ./ dVar);
 NRMSEtrainVec = sqrt(msetrainVec / dVarVec);
 NRMSEtestVec = sqrt(msetestVecresult / dVarVec);
 